;; Generated package description from bug-hunter.el  -*- no-byte-compile: t -*-
(define-package "bug-hunter" "1.3.1.0.20201128.142354" "Hunt down errors by bisecting elisp files" '((seq "1.3") (cl-lib "0.5")) :keywords '("lisp") :authors '(("Artur Malabarba" . "emacs@endlessparentheses.com")) :maintainer '("Artur Malabarba" . "emacs@endlessparentheses.com") :url "https://github.com/Malabarba/elisp-bug-hunter")
